
#define IDENTIFICATION "1997.03.17 COOKIE 2.43 <ailanto@neosplice.com>"

/* see notes pri _splitpath() and _makepath() */

/*
**
**   Authors:
**
**      William W. Patterson
**
**   History:
**
**      1995.03.11   0.00   WWP   Creation
**      1995.03.11   0.01   WWP   Set up getopt()
**      1995.03.11   0.02   WWP   Read and display the file
**      1995.03.11   0.03   WWP   Ignore comment lines (#)
**      1995.03.11   0.04   WWP   Count the noncomment lines
**      1995.03.16   0.05   WWP   Retain offset of first byte of cookie
**      1995.03.16   0.06   WWP   Let user select a cookie
**      1995.03.16   0.07   WWP   Allow continuation lines (+)
**      1995.03.17   0.08   WWP   Use display()
**      1995.03.17   1.00   WWP   Select random cookie if no index specified
**      1995.03.18   1.01   WWP   Numeric argument used instead of -#
**      1995.03.18   1.02   WWP   Added quickie option
**      1995.03.18   1.03   WWP   Do reversesearch part of quickie
**      1995.03.19   1.04   WWP   Do forwardsearch part of quickie
**      1995.03.19   1.10   WWP   Cut! Print!
**      1995.03.19   1.11   WWP   help() function split, now functions!
**      1995.03.19   1.12   WWP   Moved CookieJar default location to \toolbox\cookies\
**      1995.03.23   1.13   WWP   Include drive in default file specification
**      1995.03.24   1.14   WWP   Renamed getopt.h to getopt.c
**      1995.03.28   1.15   WWP   Move -q random offset code into a function
**      1995.03.28   1.16   WWP   Created longrand()
**      1995.03.28   1.17   WWP   Forget binary slices; use longrand() for -q
**      1995.03.28   1.18   WWP   Use longrand() everywhere, for more cookies!
**      1995.03.28   1.19   WWP   Decrement random offset if landed on newline
**      1995.03.28   1.20   WWP   Cut! Print!
**      1995.03.29   1.21   WWP   First cookie was being ignored
**      1995.03.30   1.22   WWP   Functionalize some of the main() inline code
**      1995.03.30   1.23   WWP   Functionalize some of the main() inline code
**      1995.03.30   1.24   WWP   Added -c to create index file
**      1995.03.30   1.25   WWP   Added code for index data
**      1995.03.30   1.26   WWP   Added -i to use the index file
**      1995.03.30   1.27   WWP   Damn! Forgot to include malloc.h!
**      1995.03.30   1.28   WWP   Return, don't exit, on malloc() failure
**      1995.03.30   1.29   WWP   Can't get malloc() to work. Fuckit.
**      1995.03.30   1.30   WWP   Cut! Print!
**      1995.03.30   1.31   WWP   Merge the getopt() code
**      1995.03.30   1.40   WWP   Cut! Print!
**      1995.03.30   1.41   WWP   Fixed help functions a little bit
**      1995.03.30   1.42   WWP   Traditional cookie was doing an extra linefeed
**      1995.03.30   1.43   WWP   Changed old index terminology to selection
**      1995.03.30   2.00   WWP   Cut! Print!
**      1995.09.01   2.01   WWP   Comments pri futures
**      1995.09.01   2.02   WWP   Added -p option
**      1995.09.06   2.03   WWP   Added COOKIEJAR environment variable
**      1995.09.06   2.04   WWP   Combine help functions --> help()
**      1995.09.06   2.10   WWP   Cut! Print!
**      1995.09.06   2.11   WWP   No purpose(), I added it to help()
**      1995.09.06   2.12   WWP   Added discussion of traditional cookie method
**      1995.09.06   2.20   WWP   Cut! Print!
**      1995.09.07   2.21   WWP   In help(), show selected file as well as default
**      1995.09.07   2.30   WWP   Cut! Print!
**      1995.12.06   2.31   WWP   Changed email address
**      1996.07.02   2.40   WWP   Wrap it all up for release!
**      1996.08.16   2.41   WWP   Fixed a little \n bug; see notes.
**      1996.09.20   2.42   WWP   Added note pri _splitpath() & _makepath()
**      1997.03.17   2.43   WWP   Changed email address
**
**
**   Futures:
**
**      * index automagically if -i && indexfile is older than datafile
**
**   Notes:
**
**      See help()
**
**      Default cookie selection is based on the classic method, in which the
**      entire file is read. First, cookie #1 is established as The Cookie.
**      As cookies are read, the probability of cookie #n replacing The Cookie
**      is 1/n.
**
**      The quickie version requires less disk access. A file offset is
**      selected, and the cookie containing that byte is printed. If we land
**      inside a comment, we back up to find the nearest cookie. If we hit
**      the beginning of the file without finding a cookie, search forward!
**      The major disadvantage with this method is that longer cookies are
**      more likely to be selected. However, most will be approximately the
**      the same size, oneliners, so this should not matter much.
**
**      Now there's an even faster method! Use the -c to create an index file,
**      then use -i to use it!
**
**      Assumption: there's at least one cookie in the cookiejar!
**
**      V2.41 I am reading the cookiejar in binary mode. Thus \r and \n
**      come in as separate characters. I need to ignore \r instead of
**      displaying it, or I end up with an extra carriage on the end
**      of each line. ( <CR><CR><LF> instead of <CR><LF> )
**
**      V2.42 Although I don't feel like making this change now, I discovered
**      a much simpler method for constructing the indexfilename. Use the
**      _splitpath() function to parse the cookiejar filename, substitute
**      ".i" for the extension, then pass those arguments to _makepath().
**      See TIMENAME.C for an example.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#define COOKIEJAR "c:\\toolbox\\cookies\\cookie.jar"

#define TRUE 1
#define FALSE 0

char *filename = COOKIEJAR;
FILE *CookieJar;
fpos_t FileSize = 0;
long cookies = 0;

char IndexFileName[257];
FILE *IndexFile;

time_t tod;
struct tm *now;

long selection = -1;

#define TRADITIONAL 0
#define QUICKIE 1
#define INDEXING 2
#define INDEXED 3
long function = TRADITIONAL;

short debugging = FALSE;

#define ISCOOKIE(x) ( ( x != '#' ) && ( x != '+' ) && ( x != 0x1a ) )

/******************************************************************************/

#define GETOPTERR(s,c) \
   if (opterr) \
      { \
      char errbuf[2]; \
      errbuf[0] = c; \
      errbuf[1] = '\n'; \
      (void) write(2,argv[0],(unsigned)strlen(argv[0])); \
      (void) write(2,s,(unsigned)strlen(s)); \
      (void) write(2,errbuf,2); \
      }

int opterr = 1;
int optind = 1;
int optopt;
char *optarg;

int getopt( int argc, char **argv, char *opts )

   {

   static int sp = 1;
   register int c;
   register char *cp;

   if ( sp == 1 )
      {
      if ( optind>=argc || argv[optind][0]!='-' || argv[optind][1]=='\0' )
         {
         return(EOF);
         }
      else
         {
         if ( strcmp(argv[optind],"--") == (int)NULL ) /* 002 */
            {
            ++optind;
            return(EOF);
            }
         }
      }

   optopt = c = argv[optind][sp];

   if ( c==':' || (cp=strchr(opts,c))==NULL )
      {
      GETOPTERR(": illegal option -- ", c);
      if ( argv[optind][++sp] == '\0' )
         {
         ++optind;
         sp = 1;
         }
      return('?');
      }

   if ( *++cp == ':' )
      {
      if ( argv[optind][sp+1] != '\0' )
         {
         optarg = &argv[optind++][sp+1];
         }
      else
         {
         if ( ++optind >= argc )
            {
            GETOPTERR(": option requires an argument -- ",c);
            sp = 1;
            return('?');
            }
         else
            {
            optarg = argv[optind++];
            }
         }
      sp = 1;
      }
   else
      {
      if ( argv[optind][++sp] == '\0' )
         {
         sp = 1;
         ++optind;
         }
      optarg = NULL;
      }

   return(c);

   }

/******************************************************************************/

long longrand( void )
   {
   long r;
   r = ( (long)rand() ) + ( (long)rand() << 15 );
   if ( rand() % 2 ) r |= 0x40000000;
   return r;
   }

/******************************************************************************/

void display( fpos_t *offset )

   {

   int x;
   short done = 0;
   char string[81];

   time(&tod);
   now = localtime(&tod);

   fsetpos( CookieJar, offset );

   printf("\n");

   while ( ( !done ) && ( EOF != ( x = fgetc(CookieJar) ) ) )
      {
      switch (x)
         {
         case '\r':
            break;
         case '\n':
            x = fgetc(CookieJar);
            if ( x == '+' ) printf("\n"); else done = 1;
            break;
         case '$':
            x = fgetc(CookieJar);
            switch (x)
               {
               case '\n':
                  ungetc('\n',CookieJar);
               case '$':
                  printf("$");
                  break;
               case 'B': /* bell */
                  printf("\a");
                  break;
               case 'D': /* date */
                  strftime( string, sizeof string, "%Y.%m.%d", now);
                  printf("%s",string);
                  break;
               case 'M': /* month name */
                  strftime( string, sizeof string, "%B", now);
                  printf("%s",string);
                  break;
               case 'T': /* time */
                  strftime( string, sizeof string, "%H:%M:%S", now);
                  printf("%s",string);
                  break;
               case 'W': /* weekday name */
                  strftime( string, sizeof string, "%A", now);
                  printf("%s",string);
                  break;
               case 'Y': /* year */
                  strftime( string, sizeof string, "%Y", now);
                  printf("%s",string);
                  break;
               default:
                  printf("$%c",x);
                  break;
               }
            break;
         default:
            printf("%c",x);
            break;
         }
      }

   printf("\n");

   }

/*******************************************************************************

   assumes that the filename is valid
   returns pointer to '.' or where it should go for new extension

*******************************************************************************/

char *extension( char *filename )
   {
   char *FinalBackSlash;
   char *FinalDot;
   char *TheEnd;
   TheEnd = filename + strlen(filename);
   FinalBackSlash = strrchr( filename, '\\' );
   FinalDot = strrchr( filename, '.' );
   if ( FinalDot == NULL ) return TheEnd;
   if ( FinalBackSlash == NULL ) return FinalDot;
   if ( FinalBackSlash > FinalDot ) return TheEnd;
   return FinalDot;
   }

/******************************************************************************/

void CreateIndexFile( void )

   {

   fpos_t CookieOffset = 0;
   short FirstByte = TRUE;
   int x;

   strcpy(IndexFileName,filename);
   strcpy(extension(IndexFileName),".i");
   printf("indexfilename is \"%s\"\n",IndexFileName);

   if ( NULL == ( IndexFile = fopen(IndexFileName,"wb") ) )
      {
      printf("unable to open indexfile %s\n",IndexFileName);
      }

   else

      {

      if (debugging) printf("indexfile is open\n");

      /* reserve space for cookie count, we'll fill it in later */
      fwrite(&cookies,(size_t)(sizeof cookies),(size_t)1,IndexFile);

      while ( EOF != ( x = fgetc(CookieJar) ) )
         {
         if ( FirstByte && ISCOOKIE(x) )
            {
            ++cookies;
            printf("\r%010ld: %010ld",cookies,CookieOffset);
            fwrite(&CookieOffset,(size_t)(sizeof CookieOffset),(size_t)1,IndexFile);
            }
         ++CookieOffset;
         FirstByte = ( x == '\n' );
         }

      /* it's later now; stuff in the actual cookie count */
      rewind(IndexFile);
      fwrite(&cookies,(size_t)(sizeof cookies),(size_t)1,IndexFile);

      printf("\r%ld cookies were indexed\n",cookies);

      fclose(IndexFile);

      }

   }

/******************************************************************************/

void DoIndexed( void )

   {

   fpos_t CookieOffset = 0;
   fpos_t IndexOffset = 0;

   strcpy(IndexFileName,filename);
   strcpy(extension(IndexFileName),".i");
   if (debugging) printf("indexfilename is \"%s\"\n",IndexFileName);

   if ( NULL == ( IndexFile = fopen(IndexFileName,"rb") ) )
      {
      printf("unable to open indexfile %s\n",IndexFileName);
      }
   else
      {
      if (debugging) printf("indexfile is open\n");
      fread(&cookies,(size_t)(sizeof cookies),(size_t)1,IndexFile);
      if (debugging) printf("there are %ld cookies\n",cookies);
      IndexOffset = (sizeof cookies) + (sizeof CookieOffset) * (longrand()%cookies);
      if (debugging) printf("random index offset: %ld\n",IndexOffset);
      fsetpos(IndexFile,&IndexOffset);
      fread(&CookieOffset,(size_t)(sizeof CookieOffset),(size_t)1,IndexFile);
      if (debugging) printf("cookie offset: %ld\n",CookieOffset);
      fsetpos(CookieJar,&CookieOffset);
      display(&CookieOffset);
      fclose(IndexFile);
      }

   }

/****************************************************************************/

void DoTraditional( void )

   {

   fpos_t CookieOffset = 0;
   short FirstByte = TRUE;
   int x;

   while ( ( EOF != ( x = fgetc(CookieJar) ) ) && ( cookies != selection ) )
      {
      if ( FirstByte && ISCOOKIE(x) )
         {
         ++cookies;
         if ( (cookies==selection) || (cookies==1) || (!(longrand()%cookies)) )
            {
            fgetpos(CookieJar,&CookieOffset);
            --CookieOffset;
            }
         }
      FirstByte = ( x == '\n' );
      }

   if ( debugging ) printf("%ld cookies were processed\n",cookies);

   if ( selection > cookies )
      {
      printf("\nthere's only %ld cookies in the cookiejar!\n",cookies);
      }
   else
      {
      if ( debugging ) printf("offset: %08x\n",CookieOffset);
      display(&CookieOffset);
      }

   }

/****************************************************************************/

void DoQuickie( void )

   {

   fpos_t RandomCookieOffset = 0;
   fpos_t CookieOffset = 0;
   short done = FALSE;
   short FirstByte = TRUE;
   int x;

   if (fseek(CookieJar,0L,SEEK_END))
      {
      perror("seek failed");
      }

   else

      {

      fgetpos(CookieJar,&FileSize);

      if (debugging) printf("file size is %ld\n",FileSize);

      RandomCookieOffset = longrand() % ( FileSize - 1 );

      fsetpos(CookieJar,&RandomCookieOffset);
      if ( '\n' == fgetc(CookieJar) ) --RandomCookieOffset;
      RandomCookieOffset = max(0,RandomCookieOffset); /* should not happen */

      fsetpos(CookieJar,&RandomCookieOffset);
      CookieOffset = RandomCookieOffset;
      for ( done = FALSE; CookieOffset && (!done); ) /* find cookie start */
         {
         if ( '\n' == ( x = fgetc(CookieJar) ) )
            {
            x = fgetc(CookieJar);
            done = ISCOOKIE(x);
            }
         else
            {
            --CookieOffset;
            fsetpos(CookieJar,&CookieOffset);
            }
         }
      if (CookieOffset) ++CookieOffset;
      fsetpos(CookieJar,&CookieOffset);
      x = fgetc(CookieJar);
      if ( !ISCOOKIE(x) ) /* need to do forward search */
         {
         CookieOffset = RandomCookieOffset;
         fsetpos(CookieJar,&CookieOffset);
         FirstByte = FALSE;
         done == FALSE;
         while ( (!done) && ( EOF != ( x = fgetc(CookieJar) ) ) )
            {
            if ( FirstByte && ISCOOKIE(x) )
               {
               fgetpos(CookieJar,&CookieOffset);
               --CookieOffset;
               done = TRUE;
               }
            FirstByte = ( x == '\n' );
            }
         };

      if (debugging) printf("cookie begins at %ld\n",CookieOffset);

      display(&CookieOffset);

      }

   }

/****************************************************************************/

void help( void )
   {
   printf("\n%s\n\n",IDENTIFICATION);
   printf("Default file:    %s\n",COOKIEJAR);
   printf("Selected file:   %s\n\n",filename);
   printf("Usage: cookie [options] (not case sensitive)\n");
   printf("   ? -? or -h    help\n");
   printf("   -d            show debugging info\n");
   printf("   -ffilename    select cookiejar\n");
   printf("   -q            quick cookie\n");
   printf("   -c            create index file\n");
   printf("   -i            use index file\n");
   printf("   n             show cookie #n; one only!\n\n");
   printf("Example, assuming that all files are in the current directory:\n\n");
   printf("   cookie -fcookie.jar -i\n\n");
   printf("Traditionally, this is how a cookie is selected: First, cookie #1 is\n");
   printf("named The Cookie. Then, using a random number generator, cookie #2 is\n");
   printf("given a 1/2 chance of replacing The Cookie. And so on, with cookie #n\n");
   printf("having a 1/n chance of becoming The Cookie. If you do the math, you find\n");
   printf("that all cookies have an equal chance to be The Cookie! After all cookies\n");
   printf("have been read, The Cookie is printed. The problem with the traditional\n");
   printf("method is that the entire file must be read, which could take a long time\n");
   printf("if the file is large or the system is slow.\n\n");
   printf("A quick cookie is quicker because the program simply jumps to a random\n");
   printf("offset in the cookiejar to select a cookie, rather than examining all\n");
   printf("of them in the traditional fashion. Unfortunately, longer cookies then\n");
   printf("have a higher probability of being selected.\n\n");
   printf("The quickest method is to use the -c switch to create an index file for\n");
   printf("your cookie file whenever you add cookies; then use -i when fetching\n");
   printf("a cookie. This tells the program to jump to a random offset in the index\n");
   printf("file, in which each cookie is represented by a 32 bit address, so their\n");
   printf("probabilities are once again equal.\n\n");
   printf("Instead of specifying an alternate file using -f, you can simply stuff the\n");
   printf("full filename into the environment variable COOKIEJAR. You may still override\n");
   printf("this environment variable by using the -f.\n\n");
   printf("The CookieJar is a simple ASCII textfile. Each line is a fortune cookie,\n");
   printf("but if a cookie requires more than one line, put a '+' in column one of\n");
   printf("each continuation line. Any lines with a '#' in column one are comments.\n");
   printf("The following special codes may also be embedded in the fortunes:\n");
   printf("   $$ = dollar sign\n");
   printf("   $B = beep\n");
   printf("   $D = current date\n");
   printf("   $M = current month\n");
   printf("   $T = current time\n");
   printf("   $W = current day of the week\n");
   printf("   $Y = current year\n");
   }

/****************************************************************************/


main( int argc, char **argv )

   {

   int opt;

   char *CookieEnvironmentVariable;

   union
      {
      time_t tod;
      unsigned int words[2];
      } seed;

   time(&seed.tod);
   srand(seed.words[0]);

   if ( ( CookieEnvironmentVariable = getenv("COOKIEJAR") ) != NULL )
      {
      filename = CookieEnvironmentVariable;
      }

   while ( optind < argc )
      {
      while ( ( opt = getopt(argc,argv,"?CcDdF:f:HhIiQq") ) != EOF )
         {
         switch(opt)
            {
            case 'c':
            case 'C':
               function = INDEXING;
               break;
            case 'd':
            case 'D':
               debugging = TRUE;
               break;
            case 'f':
            case 'F':
               filename = optarg;
               break;
            case 'h':
            case 'H':
            case '?':
               help();
               exit(0);
               break;
            case 'i':
            case 'I':
               function = INDEXED;
               break;
            case 'q':
            case 'Q':
               function = QUICKIE;
               break;
            default:
               break;
            }
         }
      if ( optind < argc )
         {
         switch(*argv[optind])
            {
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
               selection = atol(argv[optind]);
               selection = max( 1, selection );
               break;
            case '?':
               help();
               exit(0);
               break;
            default:
               printf("unrecognized argument: %s\n",argv[optind]);
               break;
            }
         ++optind;
         }
      }

   if (debugging) printf("filename: %s\n",filename);
   if (debugging) printf("index: %ld\n",selection);

   if ( NULL == ( CookieJar = fopen(filename,"rb") ) )
      { perror("oops"); exit(0); }

   switch (function)
      {
      case INDEXING:
         CreateIndexFile();
         break;
      case QUICKIE:
         DoQuickie();
         break;
      case TRADITIONAL:
         DoTraditional();
         break;
      case INDEXED:
         DoIndexed();
         break;
      default:
         printf("invalid function code\n");
         break;
      };

   fclose(CookieJar);

   }
